@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1997 Kungliga Tekniska Hgskolan
+ * Copyright (c) 1997, 1998, 1999 Kungliga Tekniska Hgskolan
  * (Royal Institute of Technology, Stockholm, Sweden). 
  * All rights reserved. 
  *
@@ -38,7 +38,7 @@
 
 #include "kdc_locl.h"
 
-RCSID("$Id: kaserver.c,v 1.3 1997/08/24 11:01:35 assar Exp $");
+RCSID("$Id: kaserver.c,v 1.8 1999/03/15 08:30:42 assar Exp $");
 
 #ifdef KASERVER
 
@@ -199,6 +199,8 @@ krb5_ret_xdr_data(krb5_storage *sp,
 	size_t pad = (4 - size % 4) % 4;
 
 	data->data = malloc(size);
+	if (data->data == NULL)
+	    return ENOMEM;
 	ret = sp->fetch(sp, data->data, size);
 	if(ret != size)
 	    return (ret < 0)? errno : KRB5_CC_END;
@@ -212,7 +214,7 @@ krb5_ret_xdr_data(krb5_storage *sp,
     return 0;
 }
 
-krb5_error_code
+static krb5_error_code
 krb5_store_xdr_data(krb5_storage *sp,
 		    krb5_data data)
 {
@@ -257,7 +259,6 @@ create_reply_ticket (struct rx_header *hdr,
 		     krb5_data *reply)
 {
     KTEXT_ST ticket;
-    Key *ekey = NULL;
     des_cblock session;
     krb5_storage *sp;
     krb5_data enc_data;
@@ -265,21 +266,25 @@ create_reply_ticket (struct rx_header *hdr,
     struct rx_header reply_hdr;
     des_cblock zero;
     size_t pad;
+    unsigned fyrtiosjuelva;
 
     /* create the ticket */
-    ekey = unseal_key(skey);
 
     des_new_random_key(&session);
 
     krb_create_ticket (&ticket, 0, name, instance, realm,
 		       addr->sin_addr.s_addr,
 		       &session, life, kdc_time,
-		       sname, sinstance, ekey->key.keyvalue.data);
-    hdb_free_key (ekey);
+		       sname, sinstance, skey->key.keyvalue.data);
 
     /* create the encrypted part of the reply */
     sp = krb5_storage_emem ();
+    krb5_generate_random_block(&fyrtiosjuelva, sizeof(fyrtiosjuelva));
+    fyrtiosjuelva &= 0xffffffff;
+    krb5_store_int32 (sp, fyrtiosjuelva);
+#if 0
     krb5_store_int32 (sp, 4711); /* XXX */
+#endif
     krb5_store_int32 (sp, challenge);
     sp->store  (sp, session, 8);
     memset (&session, 0, sizeof(session));
@@ -357,8 +362,10 @@ unparse_auth_args (krb5_storage *sp,
 
     krb5_ret_xdr_data (sp, &data);
     *instance = malloc(data.length + 1);
-    if (*instance == NULL)
+    if (*instance == NULL) {
+	free (*name);
 	return ENOMEM;
+    }
     memcpy (*instance, data.data, data.length);
     (*instance)[data.length] = '\0';
     krb5_data_free (&data);
@@ -419,7 +426,7 @@ do_authenticate (struct rx_header *hdr,
     }
 
     /* find a DES key */
-    ret = hdb_keytype2key(context, client_entry, KEYTYPE_DES, &ckey);
+    ret = get_des_key(client_entry, &ckey);
     if(ret){
 	kdc_log(0, "%s", krb5_get_err_text(context, ret));
 	make_error_reply (hdr, KANOKEYS, reply);
@@ -427,7 +434,7 @@ do_authenticate (struct rx_header *hdr,
     }
 
     /* find a DES key */
-    ret = hdb_keytype2key(context, server_entry, KEYTYPE_DES, &skey);
+    ret = get_des_key(server_entry, &skey);
     if(ret){
 	kdc_log(0, "%s", krb5_get_err_text(context, ret));
 	make_error_reply (hdr, KANOKEYS, reply);
@@ -435,20 +442,15 @@ do_authenticate (struct rx_header *hdr,
     }
 
     /* try to decode the `request' */
-    {
-	Key *ekey = unseal_key(ckey);
-
-	memcpy (&key, ekey->key.keyvalue.data, sizeof(key));
-	hdb_free_key(ekey);
-	des_set_key (&key, schedule);
-	des_pcbc_encrypt ((des_cblock *)request.data,
-			  (des_cblock *)request.data,
-			  request.length,
-			  schedule,
-			  &key,
-			  DES_DECRYPT);
-	memset (&schedule, 0, sizeof(schedule));
-    }
+    memcpy (&key, ckey->key.keyvalue.data, sizeof(key));
+    des_set_key (&key, schedule);
+    des_pcbc_encrypt ((des_cblock *)request.data,
+		      (des_cblock *)request.data,
+		      request.length,
+		      schedule,
+		      &key,
+		      DES_DECRYPT);
+    memset (&schedule, 0, sizeof(schedule));
 
     /* check for the magic label */
     if (memcmp ((char *)request.data + 4, "gTGS", 4) != 0) {
@@ -525,16 +527,21 @@ unparse_getticket_args (krb5_storage *sp,
 
     krb5_ret_xdr_data (sp, &data);
     *name = malloc(data.length + 1);
-    if (*name == NULL)
+    if (*name == NULL) {
+	free (*auth_domain);
 	return ENOMEM;
+    }
     memcpy (*name, data.data, data.length);
     (*name)[data.length] = '\0';
     krb5_data_free (&data);
 
     krb5_ret_xdr_data (sp, &data);
     *instance = malloc(data.length + 1);
-    if (*instance == NULL)
+    if (*instance == NULL) {
+	free (*auth_domain);
+	free (*name);
 	return ENOMEM;
+    }
     memcpy (*instance, data.data, data.length);
     (*instance)[data.length] = '\0';
     krb5_data_free (&data);
@@ -597,7 +604,7 @@ do_getticket (struct rx_header *hdr,
     }
 
     /* find a DES key */
-    ret = hdb_keytype2key(context, krbtgt_entry, KEYTYPE_DES, &kkey);
+    ret = get_des_key(krbtgt_entry, &kkey);
     if(ret){
 	kdc_log(0, "%s", krb5_get_err_text(context, ret));
 	make_error_reply (hdr, KANOKEYS, reply);
@@ -605,7 +612,7 @@ do_getticket (struct rx_header *hdr,
     }
 
     /* find a DES key */
-    ret = hdb_keytype2key(context, server_entry, KEYTYPE_DES, &skey);
+    ret = get_des_key(server_entry, &skey);
     if(ret){
 	kdc_log(0, "%s", krb5_get_err_text(context, ret));
 	make_error_reply (hdr, KANOKEYS, reply);
@@ -613,11 +620,7 @@ do_getticket (struct rx_header *hdr,
     }
 
     /* decrypt the incoming ticket */
-    {
-	Key *ekey = unseal_key(kkey);
-	memcpy (&key, ekey->key.keyvalue.data, sizeof(key));
-	hdb_free_key(ekey);
-    }
+    memcpy (&key, kkey->key.keyvalue.data, sizeof(key));
 
     /* unpack the ticket */
     {
