@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1997 Kungliga Tekniska Hgskolan
+ * Copyright (c) 1997-1999 Kungliga Tekniska Hgskolan
  * (Royal Institute of Technology, Stockholm, Sweden). 
  * All rights reserved. 
  *
@@ -37,10 +37,13 @@
  */
 
 #include "krb5_locl.h"
+#define USE_RESOLVER
+#ifdef HAVE_ARPA_NAMESER_H
+#include <arpa/nameser.h>
+#endif
+#include "resolve.h"
 
-RCSID("$Id: principal.c,v 1.31 1997/08/24 06:06:52 assar Exp $");
-
-/* Public principal handling functions */
+RCSID("$Id: principal.c,v 1.45 1999/04/24 16:42:23 joda Exp $");
 
 #define princ_num_comp(P) ((P)->name.name_string.len)
 #define princ_type(P) ((P)->name.name_type)
@@ -148,76 +151,132 @@ krb5_parse_name(krb5_context context,
     return 0;
 }
 
+static const char quotable_chars[] = "\n\t\b\\/@";
+static const char replace_chars[] = "ntb\\/@";
+
+#define add_char(BASE, INDEX, LEN, C) do { if((INDEX) < (LEN)) (BASE)[(INDEX)++] = (C); }while(0);
+
 static size_t
-quote_string(char *s, char **out)
+quote_string(const char *s, char *out, size_t index, size_t len)
 {
-    size_t len;
-    char *p;
-    char *tmp;
-    char c = 0;
-    tmp = *out;
-    if(tmp)
-	len = strlen(tmp);
-    else 
-	len = 0;
-    for(p = s; *p; p++){
-	if(*p == '\n')
-	    c = 'n';
-	else if(*p == '\t')
-	    c = 't';
-	else if(*p == '\b')
-	    c = 'b';
-	else if(*p == '\0')
-	    c = '0';
-	else if(*p == '/')
-	    c='/';      
-	else if(*p == '@')
-	    c = '@';
-	if(c){
-	    tmp = realloc(tmp, len + 2);
-	    tmp[len] = '\\';
-	    tmp[len + 1] = c;
-	    len += 2;
-	    c = 0;
-	}else{
-	    tmp = realloc(tmp, len + 1);
-	    tmp[len] = *p;
-	    len++;
-	}
+    const char *p, *q;
+    for(p = s; *p && index < len; p++){
+	if((q = strchr(quotable_chars, *p))){
+	    add_char(out, index, len, '\\');
+	    add_char(out, index, len, replace_chars[q - quotable_chars]);
+	}else
+	    add_char(out, index, len, *p);
     }
-    tmp = realloc(tmp, len + 1);
-    tmp[len] = 0;
-    *out = tmp;
-    return len;
+    if(index < len)
+	out[index] = '\0';
+    return index;
 }
 
 
-krb5_error_code
-krb5_unparse_name(krb5_context context,
-		  krb5_principal principal,
-		  char **name)
+static krb5_error_code
+unparse_name_fixed(krb5_context context,
+		   krb5_const_principal principal,
+		   char *name,
+		   size_t len,
+		   krb5_boolean short_form)
 {
-    size_t len;
-    char *s;
+    size_t index = 0;
     int i;
-    int ncomp = princ_num_comp(principal);
-    s = NULL;
-    for (i = 0; i < ncomp; i++){
-	if(i){
-	    s = realloc(s, len + 2);
-	    s[len] = '/';
-	    s[len + 1] = 0;
-	}
-	len = quote_string(princ_ncomp(principal, i), &s);
+    for(i = 0; i < princ_num_comp(principal); i++){
+	if(i)
+	    add_char(name, index, len, '/');
+	index = quote_string(princ_ncomp(principal, i), name, index, len);
+	if(index == len)
+	    return ERANGE;
+    } 
+    /* add realm if different from default realm */
+    if(short_form) {
+	krb5_realm r;
+	krb5_error_code ret;
+	ret = krb5_get_default_realm(context, &r);
+	if(ret)
+	    return ret;
+	if(strcmp(princ_realm(principal), r) != 0)
+	    short_form = 0;
+	free(r);
+    }
+    if(!short_form) {
+	add_char(name, index, len, '@');
+	index = quote_string(princ_realm(principal), name, index, len);
+	if(index == len)
+	    return ERANGE;
     }
-    s = realloc(s, len + 2);
-    s[len] = '@';
-    s[len + 1] = 0;
-    len = quote_string(princ_realm(principal), &s);
-    *name = s;
     return 0;
 }
 
+krb5_error_code
+krb5_unparse_name_fixed(krb5_context context,
+			krb5_const_principal principal,
+			char *name,
+			size_t len)
+{
+    return unparse_name_fixed(context, principal, name, len, FALSE);
+}
+
+krb5_error_code
+krb5_unparse_name_fixed_short(krb5_context context,
+			      krb5_const_principal principal,
+			      char *name,
+			      size_t len)
+{
+    return unparse_name_fixed(context, principal, name, len, TRUE);
+}
+
+static krb5_error_code
+unparse_name(krb5_context context,
+	     krb5_const_principal principal,
+	     char **name,
+	     krb5_boolean short_flag)
+{
+    size_t len = 0, plen;
+    int i;
+    krb5_error_code ret;
+    /* count length */
+    plen = strlen(princ_realm(principal));
+    if(strcspn(princ_realm(principal), quotable_chars) == plen)
+	len += plen;
+    else
+	len += 2*plen;
+    len++;
+    for(i = 0; i < princ_num_comp(principal); i++){
+	plen = strlen(princ_ncomp(principal, i));
+	if(strcspn(princ_ncomp(principal, i), quotable_chars) == plen)
+	    len += plen;
+	else
+	    len += 2*plen;
+	len++;
+    }
+    *name = malloc(len);
+    if(*name == NULL)
+	return ENOMEM;
+    ret = unparse_name_fixed(context, principal, *name, len, short_flag);
+    if(ret)
+	free(*name);
+    return ret;
+}
+
+krb5_error_code
+krb5_unparse_name(krb5_context context,
+		  krb5_const_principal principal,
+		  char **name)
+{
+    return unparse_name(context, principal, name, FALSE);
+}
+
+krb5_error_code
+krb5_unparse_name_short(krb5_context context,
+			krb5_const_principal principal,
+			char **name)
+{
+    return unparse_name(context, principal, name, TRUE);
+}
+
+#if 0 /* not implemented */
 
 krb5_error_code
 krb5_unparse_name_ext(krb5_context context,
@@ -225,10 +284,10 @@ krb5_unparse_name_ext(krb5_context context,
 		      char **name,
 		      size_t *size)
 {
-    fprintf(stderr, "krb5_unparse_name_ext: not implemented\n");
-    abort();
+    krb5_abortx(context, "unimplemented krb5_unparse_name_ext called");
 }
 
+#endif
 
 krb5_realm*
 krb5_princ_realm(krb5_context context,
@@ -316,7 +375,6 @@ build_principal(krb5_context context,
 		va_list ap)
 {
     krb5_principal p;
-    int n;
   
     p = calloc(1, sizeof(*p));
     if (p == NULL)
@@ -402,7 +460,8 @@ krb5_copy_principal(krb5_context context,
     krb5_principal p = malloc(sizeof(*p));
     if (p == NULL)
 	return ENOMEM;
-    copy_Principal(inprinc, p);
+    if(copy_Principal(inprinc, p))
+	return ENOMEM;
     *outprinc = p;
     return 0;
 }
@@ -442,6 +501,54 @@ krb5_realm_compare(krb5_context context,
     return strcmp(princ_realm(princ1), princ_realm(princ2)) == 0;
 }
 
+struct v4_name_convert {
+    const char *from;
+    const char *to; 
+} default_v4_name_convert[] = {
+    { "ftp", "ftp" },
+    { "hprop", "hprop" },
+    { "pop", "pop" },
+    { "rcmd", "host" },
+    { NULL, NULL }
+};
+
+static const char*
+get_name_conversion(krb5_context context, const char *realm, const char *name)
+{
+    struct v4_name_convert *q;
+    const char *p;
+    p = krb5_config_get_string(context, NULL, "realms", realm,
+			       "v4_name_convert", "host", name, NULL);
+    if(p == NULL)
+	p = krb5_config_get_string(context, NULL, "libdefaults", 
+				   "v4_name_convert", "host", name, NULL);
+    if(p)
+	return p;
+
+    /* XXX should be possible to override default list */
+    p = krb5_config_get_string(context, NULL,
+			       "realms",
+			       realm,
+			       "v4_name_convert",
+			       "plain",
+			       name,
+			       NULL);
+    if(p)
+	return NULL;
+    p = krb5_config_get_string(context, NULL,
+			       "libdefaults",
+			       "v4_name_convert",
+			       "plain",
+			       name,
+			       NULL);
+    if(p)
+	return NULL;
+    for(q = default_v4_name_convert; q->from; q++)
+	if(strcmp(q->from, name) == 0)
+	    return q->to;
+    return NULL;
+}
+
 krb5_error_code
 krb5_425_conv_principal_ext(krb5_context context,
 			    const char *name,
@@ -470,15 +577,11 @@ krb5_425_conv_principal_ext(krb5_context context,
 	instance = NULL;
 	goto no_host;
     }
-    p = krb5_config_get_string(context->cf, "realms", realm,
-			       "v4_name_convert", "host", name, NULL);
-    if(p == NULL)
-	p = krb5_config_get_string(context->cf, "libdefaults", 
-				   "v4_name_convert", "host", name, NULL);
+    p = get_name_conversion(context, realm, name);
     if(p == NULL)
 	goto no_host;
     name = p;
-    p = krb5_config_get_string(context->cf, "realms", realm, 
+    p = krb5_config_get_string(context, NULL, "realms", realm, 
 			       "v4_instance_convert", instance, NULL);
     if(p){
 	instance = p;
@@ -492,21 +595,38 @@ krb5_425_conv_principal_ext(krb5_context context,
 	return HEIM_ERR_V4_PRINC_NO_CONV;
     }
     if(resolve){
-	struct hostent *hp = gethostbyname(instance);
-	if(hp){
-	    instance = hp->h_name;
-	    ret = krb5_make_principal(context, &pr, 
-				      realm, name, instance, NULL);
-	    if(func == NULL || (*func)(context, pr)){
-		*princ = pr;
-		return 0;
+	const char *inst = NULL;
+#ifdef USE_RESOLVER
+	struct dns_reply *r;
+	r = dns_lookup(instance, "a");
+	if(r && r->head && r->head->type == T_A)
+	    inst = r->head->domain;
+#else
+	struct hostent *hp = roken_gethostbyname(instance);
+	if(hp)
+	    inst = hp->h_name;
+#endif
+	if(inst) {
+	    ret = krb5_make_principal(context, &pr, realm, name, inst, NULL);
+	    if(ret == 0) {
+		if(func == NULL || (*func)(context, pr)){
+		    *princ = pr;
+#ifdef USE_RESOLVER
+		    dns_free_data(r);
+#endif
+		    return 0;
+		}
+		krb5_free_principal(context, pr);
 	    }
-	    krb5_free_principal(context, pr);
 	}
+#ifdef USE_RESOLVER
+	if(r) 
+	    dns_free_data(r);
+#endif
     }
     {
 	char **domains, **d;
-	domains = krb5_config_get_strings(context->cf, "realms", realm,
+	domains = krb5_config_get_strings(context, NULL, "realms", realm,
 					  "v4_domains", NULL);
 	for(d = domains; d && *d; d++){
 	    snprintf(host, sizeof(host), "%s.%s", instance, *d);
@@ -522,7 +642,7 @@ krb5_425_conv_principal_ext(krb5_context context,
     }
     
     
-    p = krb5_config_get_string(context->cf, "realms", realm, 
+    p = krb5_config_get_string(context, NULL, "realms", realm, 
 			       "default_domain", NULL);
     if(p == NULL){
 	/* should this be an error or should it silently
@@ -539,7 +659,7 @@ krb5_425_conv_principal_ext(krb5_context context,
     krb5_free_principal(context, pr);
     return HEIM_ERR_V4_PRINC_NO_CONV;
 no_host:
-    p = krb5_config_get_string(context->cf,
+    p = krb5_config_get_string(context, NULL,
 			       "realms",
 			       realm,
 			       "v4_name_convert",
@@ -547,7 +667,7 @@ no_host:
 			       name,
 			       NULL);
     if(p == NULL)
-	p = krb5_config_get_string(context->cf,
+	p = krb5_config_get_string(context, NULL,
 				   "libdefaults",
 				   "v4_name_convert",
 				   "plain",
@@ -572,7 +692,8 @@ krb5_425_conv_principal(krb5_context context,
 			const char *realm,
 			krb5_principal *princ)
 {
-    krb5_boolean resolve = krb5_config_get_bool(context->cf, 
+    krb5_boolean resolve = krb5_config_get_bool(context,
+						NULL,
 						"libdefaults", 
 						"v4_instance_resolve", 
 						NULL);
@@ -582,32 +703,72 @@ krb5_425_conv_principal(krb5_context context,
 }
 
 
-static char*
+static int
+check_list(const krb5_config_binding *l, const char *name, const char **out)
+{
+    while(l){
+	if (l->type != krb5_config_string)
+	    continue;
+	if(strcmp(name, l->u.string) == 0) {
+	    *out = l->name;
+	    return 1;
+	}
+	l = l->next;
+    }
+    return 0;
+}
+
+static int
 name_convert(krb5_context context, const char *name, const char *realm, 
-	     const char *section)
+	     const char **out)
 {
     const krb5_config_binding *l;
-    l = krb5_config_get_list (context->cf,
+    l = krb5_config_get_list (context,
+			      NULL,
 			      "realms",
 			      realm,
 			      "v4_name_convert",
-			      section,
+			      "host",
 			      NULL);
-    if(l == NULL)
-	l = krb5_config_get_list (context->cf,
-				  "libdefaults",
-				  "v4_name_convert",
-				  section,
-				  NULL);
+    if(l && check_list(l, name, out))
+	return KRB5_NT_SRV_HST;
+    l = krb5_config_get_list (context,
+			      NULL,
+			      "libdefaults",
+			      "v4_name_convert",
+			      "host",
+			      NULL);
+    if(l && check_list(l, name, out))
+	return KRB5_NT_SRV_HST;
+    l = krb5_config_get_list (context,
+			      NULL,
+			      "realms",
+			      realm,
+			      "v4_name_convert",
+			      "plain",
+			      NULL);
+    if(l && check_list(l, name, out))
+	return KRB5_NT_UNKNOWN;
+    l = krb5_config_get_list (context,
+			      NULL,
+			      "libdefaults",
+			      "v4_name_convert",
+			      "host",
+			      NULL);
+    if(l && check_list(l, name, out))
+	return KRB5_NT_UNKNOWN;
     
-    while(l){
-	if (l->type != STRING)
-	    continue;
-	if(strcmp(name, l->u.string) == 0)
-	    return l->name;
-	l = l->next;
+    /* didn't find it in config file, try built-in list */
+    {
+	struct v4_name_convert *q;
+	for(q = default_v4_name_convert; q->from; q++) {
+	    if(strcmp(name, q->to) == 0) {
+		*out = q->from;
+		return KRB5_NT_SRV_HST;
+	    }
+	}
     }
-    return NULL;
+    return -1;
 }
 
 krb5_error_code
@@ -617,7 +778,7 @@ krb5_524_conv_principal(krb5_context context,
 			char *instance,
 			char *realm)
 {
-    char *n, *i, *r;
+    const char *n, *i, *r;
     char tmpinst[40];
     int type = princ_type(principal);
 
@@ -637,16 +798,11 @@ krb5_524_conv_principal(krb5_context context,
     }
 
     {
-	char *tmp = name_convert(context, n, r, "host");
-	if(tmp){
-	    type = KRB5_NT_SRV_HST;
+	const char *tmp;
+	int t = name_convert(context, n, r, &tmp);
+	if(t >= 0) {
+	    type = t;
 	    n = tmp;
-	}else{
-	    tmp = name_convert(context, n, r, "plain");
-	    if(tmp){
-		type = KRB5_NT_UNKNOWN;
-		n = tmp;
-	    }
 	}
     }
 
@@ -694,7 +850,7 @@ krb5_sname_to_principal (krb5_context context,
 	sname = "host";
     if(type == KRB5_NT_SRV_HST){
 	struct hostent *hp;
-	hp = gethostbyname(hostname);
+	hp = roken_gethostbyname(hostname);
 	if(hp != NULL)
 	    hostname = hp->h_name;
     }
